// SIZE: 2379

#include "ImcSongApi.h"
#include "ImcSongData.h"
#include <windows.h> 
#include <math.h>

#define PCMDATALEN (IMCSONG_LEN*16*IMCSONG_ROWLENSAMPLES)

short pcmData[PCMDATALEN] = {0,};
static WAVEHDR waveHdr;
static HWAVEOUT waveOutHandle;

static WAVEFORMATEX waveOutFormat = {
  WAVE_FORMAT_PCM,
  1,
  44100,
  88200,
  2,
  16,
  0
};

int ImcRand(int & seed) {
  seed = seed * 214013 + 2531011;
  return seed>>8;

}

int ImcIntSin(int x) {
  x = (x&255)-128;
  int ax = x>0?x:-x;
  return (x*ax>>5) - (x<<2);
}

static unsigned short ImcSongNoteFreqTable[] = {
// C      C#     D      D#     E      F      F#     G      G#     A      A#     B
  12441, 13181, 13965, 14795, 15675, 16607, 17594, 18641, 19749, 20923, 22168, 23486
};

// SIZE: 3184

static int ImcSongEnvCounterPos[IMCSONG_ENVCOUNTERLISTSIZE] = {0,};
static int ImcSongOscPosAdd[IMCSONG_OSCLISTSIZE] = {0,};
static int ImcSongOscPos[IMCSONG_OSCLISTSIZE] = {0,};
static int ImcSongOscFmPos[IMCSONG_OSCLISTSIZE] = {0,};
static int ImcSongEffectData1[IMCSONG_EFFECTLISTSIZE] = {0,};
static int ImcSongEffectData2[IMCSONG_EFFECTLISTSIZE] = {0,};
static int * ImcSongEffectHistPtr[IMCSONG_EFFECTLISTSIZE] = {NULL,};
static bool ImcSongChannelNotePlaying[8] = { false, };
static bool ImcSongChannelNoteOff[8] = { false, };

static int curSampleNum = 0;
static int curSampleNumInRow = 0;
static unsigned char curSampleNumInTick = 0;
static unsigned char curRowInPattern = 0;
static unsigned char curOrderPos = 0;

static int RenderOsc_Sine(unsigned int pos) {
  return ImcIntSin(pos>>8)*256;
}

static int RenderOsc_Saw(unsigned int pos) {
  return pos-32768;
}

static int RenderOsc_Square(unsigned int pos) {
  return (pos>32767)?32767:-32768;
}

static int RenderOsc_Noise(unsigned int pos) {
  static int rndSeed;
  return (ImcRand(rndSeed)&65535)-32768;
}

typedef int (*TRenderOscFn)(unsigned int);

TRenderOscFn OscRenderers[] = { RenderOsc_Sine, RenderOsc_Saw, RenderOsc_Square, RenderOsc_Noise };

static int RenderSample() {
  int channelSample[8] = {0, };
  static int rndSeed;

  for(int i=0;i<IMCSONG_OSCLISTSIZE;i++) {
    TImcSongOscillator * osc = &ImcSongOscillatorList[i];
    int oscVal = OscRenderers[osc->type](ImcSongOscPos[i]&65535);
    int vol = (osc->vol * ImcSongEnvCounterList[osc->volEnvCounterId].val) >> 8;
    oscVal = (oscVal * vol) >> 8;
    if(osc->fmTargetOscId<255) {
      ImcSongOscFmPos[osc->fmTargetOscId] = oscVal;
    } else {
      if(ImcSongChannelNotePlaying[osc->channel])
        channelSample[osc->channel] += oscVal;
    }
  }

  int s = 0;
  for(int channel=0;channel<8;channel++) {

    int channelVol = (ImcSongChannelVol[channel] * ImcSongEnvCounterList[ImcSongChannelEnvCounter[channel]].val) >> 8;

    int inSample = (channelSample[channel] * channelVol) >> 8;
    int outSample = inSample;

    for(int i=0;i<IMCSONG_EFFECTLISTSIZE;i++) {
      if(ImcSongEffectList[i].channel != channel)
        continue;

      int varSample1 = ImcSongEffectData1[i];
      int varSample2 = ImcSongEffectData2[i];
      int fxSample1 = ImcSongEffectList[i].v1;
      int fxSample2 = ImcSongEffectList[i].v2;
      int fxEnvSample1 = (fxSample1 * ImcSongEnvCounterList[ImcSongEffectList[i].envCounterId1].val) >> 8;
      int fxEnvSample2 = (fxSample2 * ImcSongEnvCounterList[ImcSongEffectList[i].envCounterId2].val) >> 8;

      int * histPtr = ImcSongEffectHistPtr[i];
      int histSize = ImcSongEffectList[i].histSize;
      int histPos = curSampleNum % histSize;
      int histSample = histPtr[histPos];

      switch(ImcSongEffectList[i].type) {

#ifdef IMCSONG_ENABLE_DELAYEFFECT
        case IMCSONGEFFECTTYPE_DELAY:
          outSample =
            inSample +
            ((histSample * fxSample1) >> 8);
          histSample = outSample;
          break;
#endif

#ifdef IMCSONG_ENABLE_FLANGEEFFECT
        case IMCSONGEFFECTTYPE_FLANGE:
          {
            int histReadPos =
              (
                histPos -
                ImcSongEnvCounterList[ImcSongEffectList[i].envCounterId1].val +
                histSize
              ) % histSize;
            outSample =
              inSample +
              histPtr[histReadPos];
            histSample = inSample;
          }
          break;
#endif

#ifdef IMCSONG_ENABLE_LOWPASSEFFECT
        case IMCSONGEFFECTTYPE_LOWPASS:
          outSample = (inSample*fxEnvSample1)/256 + (histSample*(255-fxEnvSample1))/256;
          histSample = outSample;
          break;
#endif

#ifdef IMCSONG_ENABLE_HIGHPASSEFFECT
        case IMCSONGEFFECTTYPE_HIGHPASS:
          {
            int invFxEnvSample1 = 255 - fxEnvSample1;
            outSample =
              ( 
                inSample*((256+invFxEnvSample1)/2) +
                varSample1*(-(256+invFxEnvSample1)/2) +  
                varSample2*invFxEnvSample1
              ) >> 8;
            varSample1 = inSample;
            varSample2 = outSample;
          }
          break;
#endif

#ifdef IMCSONG_ENABLE_RESONANCEEFFECT
        case IMCSONGEFFECTTYPE_RESONANCE:
          {
            int ifc = fxEnvSample1;
            int ifr = 255 - fxEnvSample2;
            int iff = 255 - (ifc*ifr)/256;
            varSample1 = (iff*varSample1 - ifc*varSample2 + ifc*inSample) >> 8;
            varSample2 = (iff*varSample2 + ifc*varSample1) >> 8;
            outSample = varSample2;
          }
          break;
#endif

#ifdef IMCSONG_ENABLE_OVERDRIVEEFFECT
        case IMCSONGEFFECTTYPE_OVERDRIVE:
          if(inSample>fxSample1)
            inSample = fxSample1;
          if(inSample<-fxSample1)
            inSample = -fxSample1;
          outSample = (inSample*fxEnvSample2) >> 8;
          break;
#endif
      }

      ImcSongEffectData1[i] = varSample1;
      ImcSongEffectData2[i] = varSample2;
      histPtr[histPos] = histSample;
      inSample = outSample;

    }

    s += (outSample * IMCSONG_VOL) >> 8;

  }
  
  if(s<-32768) s=-32768;
  if(s>32767) s=32767;
  return s;
}

static void RowHit() {
  if(curOrderPos>=IMCSONG_LEN)
    return;

  for(int channel=0;channel<8;channel++) {
    int pattern = (ImcSongOrderTable[curOrderPos] >> (channel*4)) & 15;
    if(pattern==0)
      continue;

    int patternIndex = ((int)ImcSongPatternLookupTable[channel] + pattern-1) * 16;
    unsigned char * pptr = ImcSongPatternData+patternIndex;
    int note = pptr[curRowInPattern];
    
    if(note==0)
      continue;
    
    if(note==255) {
      ImcSongChannelNoteOff[channel] = true;
      if(ImcSongChannelStopNote[channel])
        ImcSongChannelNotePlaying[channel] = false;
      continue;
    }
    
    ImcSongChannelNoteOff[channel] = false;
    ImcSongChannelNotePlaying[channel] = true;
    
    for(int i=0;i<IMCSONG_ENVCOUNTERLISTSIZE;i++) {
      if(ImcSongEnvCounterList[i].channel != channel)
        continue;
      int envId = ImcSongEnvCounterList[i].envId;
      if(envId==255)
        continue;
      if(ImcSongEnvList[envId].resetOnNewNote)
        ImcSongEnvCounterPos[i] = 0;
    }

    for(int i=0;i<IMCSONG_OSCLISTSIZE;i++) {
      if(ImcSongOscillatorList[i].channel != channel)
        continue;
      ImcSongOscPos[i] = 0;
      ImcSongOscFmPos[i] = 0;
      int octave = note>>4;
      int noteIndex = note&15;
      int posAdd = ((unsigned int)ImcSongNoteFreqTable[noteIndex] * (1<<octave)) / 512;
      posAdd = (posAdd << ImcSongOscillatorList[i].transOctave) >> 8;
      posAdd = (posAdd * (ImcSongOscillatorList[i].transFine + 256)) >> 8;
      ImcSongOscPosAdd[i] = posAdd;
    }
  }

}

static void TickHit() {
  for(int i=0;i<IMCSONG_ENVCOUNTERLISTSIZE;i++) {
    int channel = ImcSongEnvCounterList[i].channel;
    int envId = ImcSongEnvCounterList[i].envId;
    if(envId==255)
      continue;
    int pos = ImcSongEnvCounterPos[i] + ImcSongEnvList[envId].speed;
    
    if(
      !ImcSongChannelNoteOff[channel] &&
      ImcSongEnvList[envId].sustain<255 &&
      pos > ImcSongEnvList[envId].sustain*2048
    )
      pos = ImcSongEnvList[envId].sustain*2048;

    if(ImcSongEnvList[envId].keep<255 && pos > ImcSongEnvList[envId].keep*2048)
      pos = ImcSongEnvList[envId].keep*2048;

    ImcSongEnvCounterList[i].val = 
      (ImcIntSin((pos+ImcSongEnvList[envId].phase*2048)>>8) + 128) *
      (ImcSongEnvList[envId].maxVal-ImcSongEnvList[envId].minVal) / 256 +
      ImcSongEnvList[envId].minVal;

    ImcSongEnvCounterPos[i] = pos;
  }
}

static void Advance() {
  curSampleNum++;
  curSampleNumInRow++;
  if(curSampleNumInRow>=IMCSONG_ROWLENSAMPLES) {
    curSampleNumInRow=0;
    curRowInPattern++;
    if(curRowInPattern>15) {
      curRowInPattern=0;
      curOrderPos++;
      //-- Pattern hit
      for(int i=0;i<IMCSONG_ENVCOUNTERLISTSIZE;i++) {
        int envId = ImcSongEnvCounterList[i].envId;
        if(envId==255)
          continue;
        if((curOrderPos&ImcSongEnvList[envId].resetOnPatternMask)==0)
          ImcSongEnvCounterPos[i] = 0;
      }
      //--
    }
    RowHit();
  }
  curSampleNumInTick++;
  if(curSampleNumInTick>=44) {
    curSampleNumInTick=0;
    TickHit();
  }
  for(int i=0;i<IMCSONG_OSCLISTSIZE;i++) {
    int posAdd =
      ((ImcSongOscPosAdd[i] *  
      ImcSongEnvCounterList[ImcSongOscillatorList[i].transEnvCounterId].val) >> 8) +  
      (ImcSongOscFmPos[i] >> 3);
    ImcSongOscPos[i] += posAdd;
  }
}

static DWORD CALLBACK BufferFillerThread(void * param) {
  for(int i=0;i<PCMDATALEN;++i) {
    pcmData[i] = RenderSample();
    Advance();
  }
  return 0;
}

/////////////////////////////////////////////////////////////////////////

static int songStartTime;

void ImcSongPlay() {
  for(int i=0;i<IMCSONG_EFFECTLISTSIZE;i++) {
    ImcSongEffectHistPtr[i] = (int *)LocalAlloc(LPTR, ImcSongEffectList[i].histSize*4);
  }
  RowHit();
  CreateThread(NULL,0,BufferFillerThread,NULL,0,NULL);
  waveOutOpen(&waveOutHandle, WAVE_MAPPER, &waveOutFormat, NULL, 0, CALLBACK_NULL);
  waveHdr.lpData = (LPSTR)pcmData;
  waveHdr.dwBufferLength = IMCSONG_LEN*16*IMCSONG_ROWLENSAMPLES*2;
  waveHdr.dwFlags = 0;
  waveOutPrepareHeader(waveOutHandle, &waveHdr, sizeof(WAVEHDR));
  waveOutWrite(waveOutHandle, &waveHdr, sizeof(WAVEHDR));
  songStartTime = GetTickCount();
}

void ImcSongUpdatePos(TImcSongPos & pos) {
  pos.curTime = GetTickCount() - songStartTime;
  int sample = (441 * pos.curTime) / 10;
  int row = sample / IMCSONG_ROWLENSAMPLES;
  pos.curOrderPos = row / 16;
  pos.curRowInPattern = row & 15;
  pos.curSampleInRow = sample % IMCSONG_ROWLENSAMPLES;
  pos.curSample = sample;
}

